/**
 * This file is distributed as part of the code accompanying the paper:
 *
 * "Robust Region Detection via Consensus Segmentation of Deformable Shapes"
 * E. Rodola, S. Rota Bulo, and D. Cremers
 * Computer Graphics Forum, 33(5), 2014
 *
 * If you use this code, please cite the paper above.
 *
 * Samuel Rota Bulò (c) 2014
 */
#ifndef __CONSENSUS_H__
#define __CONSENSUS_H__

#include <iostream>

#ifndef NDEBUG
namespace std
{
void *myMalloc(size_t size)
{
// log
  std::cout << "Malloc " <<size << std::endl;
  return malloc(size);
}

void myFree(void *ptr)
{
// log
  free(ptr);
}
}

#define malloc(size) myMalloc(size)
#define free(ptr) myFree(ptr)
#endif

using namespace std;
#include <Eigen/Core>
#include <Eigen/Dense>



using namespace Eigen;

namespace cvlab
{

void PPC_l2(const Ref<const MatrixXd> C,const Ref<const MatrixXd> N, Ref<MatrixXd> Y,int k, double toll,int maxIter, int& niter, double& final_err)
{

  int n=C.rows();

  /* Y=rand(k,n)+1000; Y=Y./repmat(sum(Y),k,1); */
  Y.setRandom();
  Y.array()+=1000;
  Y.array().rowwise()/=Y.colwise().sum().array(); //column-wise normalization

  /* A=(N.*(Y'Y)-C)+diag(0.5*diag(N)-diag(C)); */
  MatrixXd A=MatrixXd::Zero(n,n);
  A.noalias()=Y.transpose()*Y;
  A=A.array()*N.array()-C.array();
  A.diagonal()=0.5*N.diagonal()-C.diagonal();

  /* G=Y*A */
  MatrixXd G=MatrixXd::Zero(k,n);
  G.noalias()=-Y*A;

  bool notConverged=true;

//  VectorXi vMax=VectorXi(n);
//  VectorXi vMin=VectorXi(n);

  niter=0;
  double bestGap;

  while(notConverged && (maxIter<=0 || niter<maxIter)) {
    /* determine search direction */
    int simplexIdx=0;
    int maxIdx=-1, minIdx=-1;
    bestGap=-1;

//    #pragma omp parallel for
//    for(int i=0; i<n; ++i) {
//      int lMaxIdx=0, lMinIdx=-1;
//      for(int u=0; u<k; ++u) {
//        if(G(u,i)>G(lMaxIdx,i))
//          lMaxIdx=u;
//        if(Y(u,i)>0 && (lMinIdx<0 || G(u,i)<G(lMinIdx,i)))
//          lMinIdx=u;
//      }
//      vMax(i)=lMaxIdx;
//      vMin(i)=lMinIdx;
//    }
//
//    for(int i=0; i<n; ++i) {
//      double curGap=G(vMax(i),i)-G(vMin(i),i);
//      if(bestGap<curGap) {
//        bestGap=curGap;
//        simplexIdx=i;
//      }
//    }
//    maxIdx=vMax(simplexIdx);
//    minIdx=vMin(simplexIdx);


    for(int i=0; i<n; ++i) {
      int lMaxIdx=0, lMinIdx=-1;
      for(int u=0; u<k; ++u) {
        if(G(u,i)>G(lMaxIdx,i))
          lMaxIdx=u;
        if(Y(u,i)>0 && (lMinIdx<0 || G(u,i)<G(lMinIdx,i)))
          lMinIdx=u;
      }
      if(G(lMaxIdx,i)-G(lMinIdx,i)>bestGap) {
        simplexIdx=i;
        maxIdx=lMaxIdx;
        minIdx=lMinIdx;
        bestGap=G(maxIdx,simplexIdx)-G(minIdx,simplexIdx);
      }
    }
  
    //cout << bestGap << " " << simplexIdx<<" " << maxIdx << " " << minIdx << " " << Y(minIdx,simplexIdx) << endl;

    notConverged=bestGap>toll;
    if(notConverged) {
      /* determine optimal step size */
      double den=(N.row(simplexIdx).array()*((Y.row(maxIdx)-Y.row(minIdx)).array().square())).sum();
      den-=N(simplexIdx,simplexIdx)*(Y(maxIdx,simplexIdx)-Y(minIdx,simplexIdx))*(Y(maxIdx,simplexIdx)-Y(minIdx,simplexIdx));
      den+=N(simplexIdx,simplexIdx)-2*C(simplexIdx,simplexIdx);
      double alpha=den<=0?Y(minIdx,simplexIdx):min(bestGap/den,Y(minIdx,simplexIdx));

      //cout << bestGap <<" " << alpha << endl;

      /*update G (undo)*/
      G.noalias()+=Y.col(simplexIdx)*A.row(simplexIdx);

      /*update A*/
      double tmp=A(simplexIdx,simplexIdx);
      A.row(simplexIdx).array()+=alpha*N.row(simplexIdx).array()*(Y.row(maxIdx)-Y.row(minIdx)).array();
      for(int i=0; i<n; ++i)A(i,simplexIdx)=A(simplexIdx,i);
      A(simplexIdx,simplexIdx)=tmp;

      /*update Y*/
      Y(maxIdx,simplexIdx)+=alpha;
      Y(minIdx,simplexIdx)-=alpha;

      /*update G (redo)*/
      G.noalias()-=Y.col(simplexIdx)*A.row(simplexIdx);
      G.col(simplexIdx).noalias()=-Y*A.row(simplexIdx).transpose();

      /* FOR DEBUG */
//      cout << G << endl;
//      cout << "****" << endl;
//
//  MatrixXd At=MatrixXd::Zero(n,n);
//  At.noalias()=Y.transpose()*Y;
//  At=At.array()*N.array()-C.array();
//  At.diagonal()=0.5*N.diagonal()-C.diagonal();
//
//  /* G=Y*A */
//  MatrixXd Gt=MatrixXd::Zero(k,n);
//  Gt.noalias()=-Y*At;
//    cout << Gt << endl;

      /* PRINT OBJECTIVE - SLOW */
//      if(niter%1000 == 0) {
//        double obj=0;
//        for(int i=0; i<n; ++i)
//          for(int j=0; j<n; ++j)
//            if(i!=j)
//              obj+=N(i,j)*pow(Y.col(i).transpose()*Y.col(j)-C(i,j)/N(i,j),2);
//            else
//              obj+=Y.col(i).squaredNorm()*(N(i,i)-2*C(i,i));
//        cout << obj << endl;
//      }
      ++niter;
    }
	
  }
  //cout << "niter=" << niter << endl;
  final_err = bestGap;
}


}


#endif
